@@ -91,6 +91,14 @@ fn main() {
             .open(tbf_path.clone())
             .unwrap();
 
+        // Adding padding to the end of cortex-m apps. Check for a cortex-m app
+        // by inspecting the "machine" value in the elf header. 0x28 is ARM (see
+        // https://en.wikipedia.org/wiki/Executable_and_Linkable_Format#File_header
+        // for a list).
+        //
+        // RISC-V apps do not need to be sized to power of two.
+        let add_trailing_padding = elffile.ehdr.machine.0 == 0x28;
+
         // Do the conversion to a tock binary.
         if opt.verbose {
             println!("Creating {:?}", tbf_path);
@@ -106,6 +114,7 @@ fn main() {
             opt.protected_region_size,
             opt.permissions.to_vec(),
             minimum_tock_kernel_version,
+            add_trailing_padding,
         )
         .unwrap();
         if opt.verbose {
@@ -116,12 +125,6 @@ fn main() {
         outfile.seek(io::SeekFrom::Start(0)).unwrap();
         tab.append_file(tbf_path.file_name().unwrap(), &mut outfile)
             .unwrap();
-        outfile.seek(io::SeekFrom::Start(0)).unwrap();
-        tab.append_file(
-            tbf_path.with_extension("bin").file_name().unwrap(),
-            &mut outfile,
-        )
-        .unwrap();
     }
 }
 
@@ -148,6 +151,7 @@ fn elf_to_tbf<W: Write>(
     protected_region_size_arg: Option<u32>,
     permissions: Vec<(u32, u32)>,
     kernel_version: Option<(u16, u16)>,
+    trailing_padding: bool,
 ) -> io::Result<()> {
     let package_name = package_name.unwrap_or_default();
 
@@ -446,7 +450,22 @@ fn elf_to_tbf<W: Write>(
 
     let mut entry_point_found = false;
 
-    // Iterate the sections in the ELF file and add them to the binary as needed
+    // We need to keep track of the address in the elf file for each section we
+    // are adding to the binary. Sections can have padding between them that we
+    // need to preserve. So, we track where the last section we wrote to our
+    // output binary ended in the segment address space specified in the .elf.
+    //
+    // *********
+    // TODO! (added 08/2021, but known about for a while before that)
+    // *********
+    // elf2tab needs to be re-written to use segments rather the hack we have
+    // here. We just assume we can ad-hoc determine the sections we need to
+    // include, when we should just use the segment mapping. Because of this,
+    // `last_section_address_end` is a hack as well.
+    let mut last_section_address_end: Option<usize> = None;
+
+    // Iterate the sections in the ELF file. The sections are sorted in order of
+    // offset. Add the sections we need to the binary.
     for s in &sections_sort {
         let section = &input.sections[s.0];
 
@@ -479,6 +498,36 @@ fn elf_to_tbf<W: Write>(
             && section.shdr.shtype == elf::types::SHT_PROGBITS
             && section.shdr.size > 0
         {
+            // This is a section we are going to add to the binary.
+
+            if last_section_address_end.is_some() {
+                // We have a previous section. Now, check if there is any
+                // padding between the sections in the .elf.
+                let end = last_section_address_end.unwrap();
+                let start = section.shdr.addr as usize;
+
+                // Because we have flash and ram memory regions, we have
+                // multiple address spaces. This check lets us assume we are in
+                // a new address segment. We need the start of the next section
+                // to be after the previous one, and the gap to not be _too_
+                // large.
+                if start > end && (start - end) < 100 {
+                    // If this is the next section in the same segment, then
+                    // check if there is any padding required.
+                    let padding = start - end;
+                    if padding > 0 {
+                        if verbose {
+                            println!("  Adding {} bytes of padding between sections", padding,);
+                        }
+
+                        // Increment our index pointer and add the padding bytes.
+                        binary_index += padding;
+                        let zero_buf = [0_u8; 512];
+                        binary.extend(&zero_buf[..padding]);
+                    }
+                }
+            }
+
             if verbose {
                 println!(
                     "  Adding {0} section. Offset: {1} ({1:#x}). Length: {2} ({2:#x}) bytes.",
@@ -506,6 +555,9 @@ fn elf_to_tbf<W: Write>(
 
             // Now increment where we are in the binary.
             binary_index += section.shdr.size as usize;
+
+            // And update our end in the .elf offset address space.
+            last_section_address_end = Some((section.shdr.addr + section.shdr.size) as usize);
         }
     }
 
@@ -553,16 +605,25 @@ fn elf_to_tbf<W: Write>(
     // the relocation data length.
     binary_index += relocation_binary.len() + mem::size_of::<u32>();
 
-    // That is everything that we are going to include in our app binary. Now
-    // we need to pad the binary to a power of 2 in size, and make sure it is
-    // at least 512 bytes in size.
-    let post_content_pad = if binary_index.count_ones() > 1 {
-        let power2len = cmp::max(1 << (32 - (binary_index as u32).leading_zeros()), 512);
-        power2len - binary_index
+    // That is everything that we are going to include in our app binary.
+
+    let post_content_pad = if trailing_padding {
+        // If trailing padding is requested, we need to pad the binary to a
+        // power of 2 in size, and make sure it is at least 512 bytes in size.
+        let pad = if binary_index.count_ones() > 1 {
+            let power2len = cmp::max(1 << (32 - (binary_index as u32).leading_zeros()), 512);
+            power2len - binary_index
+        } else {
+            0
+        };
+        // Increment to include the padding.
+        binary_index += pad;
+        pad
     } else {
+        // No padding.
         0
     };
-    binary_index += post_content_pad;
+
     let total_size = binary_index;
 
     // Now set the total size of the app in the header.
@@ -580,7 +641,7 @@ fn elf_to_tbf<W: Write>(
     output.write_all(&rel_data_len)?;
     output.write_all(relocation_binary.as_ref())?;
 
-    // Pad to get a power of 2 sized flash app.
+    // Pad to get a power of 2 sized flash app, if requested.
     util::do_pad(output, post_content_pad as usize)?;
 
     Ok(())
